/*
 * Copyright 2008 Kyoto University and Honda Motor Co.,Ltd.
 * All rights reserved.
 * HARK was developed by researchers in Okuno Laboratory
 * at the Kyoto University and Honda Research Institute Japan Co.,Ltd.
 */

#include <iostream>
#include <BufferedNode.h>
#include <Buffer.h>
#include <Matrix.h>

using namespace FD;
using namespace std;

class MatrixToMatrix;

DECLARE_NODE(MatrixToMatrix)
/*Node
 *
 * @name MatrixToMatrix
 * @category HARK:MISC
 * @description Convert ObjectRef of Matrix<ObjectRef>. If the input is Matrix<float>, the output is Matrix<complex<float> >. If the input is Matrix<complex<float> >, the output is Matrix<float>.
 *
 * @input_name INPUT
 * @input_type any
 * @input_description Matrix<float> or Matrix<complex<float> >.
 *
 * @output_name OUTPUT
 * @output_type any	
 * @output_description Matrix<float> or Matrix<complex<float> >.
 *
 * @parameter_name METHOD_COMPLEX_TO_FLOAT
 * @parameter_type string
 * @parameter_list magnitude:real:imaginary
 * @parameter_value magnitude
 * @parameter_description Method og converting complex to float. This is used only when INPUT is Matrix<complex<float> >. [default: magnitude]
 *
 * @parameter_name METHOD_FLOAT_TO_COMPLEX
 * @parameter_type string
 * @parameter_list zero:hilbert
 * @parameter_value zero
 * @parameter_description Method of converting float to complex. This is used only when INPUT is Matrix<float>. [default: zero]
 *
 * @parameter_name DEBUG
 * @parameter_type bool
 * @parameter_list true:false
 * @parameter_value false
 * @parameter_description Show debug points. [default: false]
 *
 END*/

static const int COMPLEX_ZERO = 0;
static const int COMPLEX_HILBERT = 1;

static const int SOURCE_MAGNITUDE = 0;
static const int SOURCE_REAL = 1;
static const int SOURCE_IMAGINARY = 2;

class MatrixToMatrix : public BufferedNode {
	int inputID;
	int outputID;
	int method;
	int complex_method;
	bool debug;
    
public:
	MatrixToMatrix(string nodeName, ParameterSet params)
		: BufferedNode(nodeName, params) {
		inputID = addInput("INPUT");
		outputID = addOutput("OUTPUT");
		string method_str = object_cast<String>(parameters.get("METHOD_COMPLEX_TO_FLOAT"));
		if (method_str == "magnitude")      method = SOURCE_MAGNITUDE;
		else if (method_str == "real")      method = SOURCE_REAL;
		else if (method_str == "imaginary") method = SOURCE_IMAGINARY;
		else {
			cerr << "******* MatrixToMatrix: Input error" << endl;
			throw new NodeException(NULL, "MatrixToMatrix: METHOD_COMPLEX_TO_FLOAT should be \"magnitude\" or \"real\" or \"imaginary\".", __FILE__, __LINE__);
		}
		string complex_str = object_cast<String>(parameters.get("METHOD_FLOAT_TO_COMPLEX"));
		if (complex_str == "zero")         complex_method = COMPLEX_ZERO;
		else if (complex_str == "hilbert") complex_method = COMPLEX_HILBERT;
		else {
			cerr << "******* MatrixToMatrix: Input error" << endl;
			throw new NodeException(NULL, "MatrixToMatrix: METHOD_FLOAT_TO_COMPLEX should be \"zero\" or \"hilbert\".", __FILE__, __LINE__);
		}
		debug = dereference_cast<bool>(parameters.get("DEBUG"));
	}

	void calculate(int output_id, int count, Buffer &out) {
		ObjectRef input = getInput(inputID, count);

		if (debug == true) {
			cout << "******* MatrixToMatrix debug point-Input check (count=" << count << ")" << endl;
            cout << "Input ObjectRef's type is: ";
			if (typeid(*input) == typeid(Matrix<float>))                cout << "Matrix<float>" << endl;
			else if (typeid(*input) == typeid(Matrix<complex<float> >)) cout << "Matrix<complex<float> >" << endl;
			else                                                        cout << "unidentified." << endl;
		}

		if (typeid(*input) == typeid(Matrix<float>)) {
			const Matrix<float> &mtFltInput = object_cast<Matrix<float> >(input);

			RCPtr<Matrix<complex<float> > > output(new Matrix<complex<float> >(mtFltInput.nrows(), mtFltInput.ncols()));
			out[count] = output;

			if (complex_method == COMPLEX_HILBERT) {
				for (int row = 0; row < mtFltInput.nrows(); row++) {
					for (int col = 0; col < mtFltInput.ncols(); col++) {
						(*output)(row, col) = complex<float>(mtFltInput(row, col), abs(mtFltInput(row, col)));
					}
				}
			}
			else { // complex_method == COMPLEX_ZERO
				for (int row = 0; row < mtFltInput.nrows(); row++) {
					for (int col = 0; col < mtFltInput.ncols(); col++) {
						(*output)(row, col) = complex<float>(mtFltInput(row, col), 0.0f);
					}
				}
			}
		}
		else if (typeid(*input) == typeid(Matrix<complex<float> >)) {
			const Matrix<complex<float> > &mtCmpInput = object_cast<Matrix<complex<float> > >(input);

			RCPtr<Matrix<float> > output(new Matrix<float>(mtCmpInput.nrows(), mtCmpInput.ncols()));
			out[count] = output;
	
			if (method == SOURCE_REAL) {
				for (int row = 0; row < mtCmpInput.nrows(); row++) {
					for (int col = 0; col < mtCmpInput.ncols(); col++) {
						(*output)(row, col) = mtCmpInput(row, col).real();
					}
				}
			}
			else if (method == SOURCE_IMAGINARY) {
				for (int row = 0; row < mtCmpInput.nrows(); row++) {
					for (int col = 0; col < mtCmpInput.ncols(); col++) {
						(*output)(row, col) = mtCmpInput(row, col).imag();
					}
				}
			}
			else { // method == SOURCE_MAGNITUDE
				for (int row = 0; row < mtCmpInput.nrows(); row++) {
					for (int col = 0; col < mtCmpInput.ncols(); col++) {
						(*output)(row, col) = abs(mtCmpInput(row, col));
					}
				}
			}	
		}
		else {
			cerr << "******* MatrixToMatrix: Input type is unidentified" << endl;
			throw(new NodeException(NULL, "Exception caught in calculate[MatrixToMatrix]", __FILE__, __LINE__));
		}

        if (debug == true ) cout << "******* MatrixToMatrix debug point-Finish" << endl;
    }

    NO_ORDER_NODE_SPEEDUP(MatrixToMatrix);
};
